use std::{collections::HashMap, error::Error, fmt::Display};

use crate::{
    handling::{methods::Method, routes::Uri},
    utils::mime::mime_enum::Mime,
};

pub trait FromRequest: Send {
    fn get_data(&self) -> &Self;
    fn set_data(&mut self, data: &Self);
    fn append(&mut self, data: &Self);
}

impl FromRequest for Vec<u8> {
    fn get_data(&self) -> &Self {
        &self
    }

    fn set_data(&mut self, data: &Self) {
        *self = data.to_vec();
    }
    fn append(&mut self, data: &Self) {
        self.extend_from_slice(data);
    }
}

type HeaderMap = Vec<String>;

#[derive(Clone)]
pub struct Request<'a> {
    pub uri: Uri<'a>,
    pub headers: HeaderMap,
    pub method: Method,
    pub cookies: Option<HashMap<String, String>>,
    pub mime_type: Option<Mime>,
    // pub connection: ConnectionMeta,
}

// struct ConnectionMeta {
//     remote: Option<SocketAddr>,
//     // certificates
// }

#[derive(Clone, Debug, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum MediaType {
    Json,
    Plain,
    Html,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum ParseErrors {
    NoData,
    BadData,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct ParseFormError {
    pub error: ParseErrors,
}

impl Display for ParseFormError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.error)
    }
}

impl Display for ParseErrors {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ParseErrors::NoData => write!(f, "No Data at key"),
            ParseErrors::BadData => write!(f, "Bad Data at key"),
        }
    }
}

impl Error for ParseFormError {}
